@@ -229,7 +229,49 @@ class FLEBasis2D:
             nus[2 * i] = i
         c2r_nus = self.precomp_transform_complex_to_real(nus)
         r2c_nus = spr.csr_matrix(c2r_nus.transpose().conj())
+        
+        xs = 1 - (2 * np.arange(n_radial) + 1) / (2 * n_radial)
+        xs = np.cos(np.pi * xs)
+        pts = (xs + 1) / 2
+        pts = (lmd1 - lmd0) * pts + lmd0
+        pts = pts.reshape(-1, 1)
 
+
+        blk_size = np.zeros(2 * nmax + 1)
+        for i in range(2 * nmax + 1):
+            blk_size[i] = len(idx_list[i])
+
+        blk_size = blk_size.astype(int)
+        k_max = blk_size[::2]
+        blk_ind = np.concatenate((np.zeros(1), np.cumsum(blk_size)))
+        blk_ind = blk_ind.astype(int)
+        n_blk = len(blk_ind) - 1
+        indices_sgns = np.zeros(ne, dtype=int)
+        indices_ells = np.zeros(ne, dtype=int)
+        i = 0
+        ci = 0
+        for ell in range(nmax + 1):
+            sgns = (1,) if ell == 0 else (1, -1)
+            ks = np.arange(0, k_max[ell])
+
+            for sgn in sgns:
+                rng = np.arange(i, i + len(ks))
+                indices_ells[rng] = ell
+                indices_sgns[rng] = sgn
+
+                i += len(ks)
+
+            ci += len(ks)
+
+
+
+        self.indices_sgns = indices_sgns
+        self.indices_ells = indices_ells
+        self.pts = pts
+        self.blk_ind = blk_ind
+        self.k_max = k_max
+        self.n_blk = n_blk
+        self.blk_size = blk_size
         self.ns = ns
         self.ks = ks
         self.lmds = lmds
@@ -262,10 +304,7 @@ class FLEBasis2D:
         if L < 16:
             self.B = self.create_denseB()
 
-        xs = 1 - (2 * np.arange(self.n_radial) + 1) / (2 * self.n_radial)
-        xs = np.cos(np.pi * xs)
-        pts = (xs + 1) / 2
-        pts = (self.lmd1 - self.lmd0) * pts + self.lmd0
+
         R = L // 2
         h = 1 / R
         phi = 2 * np.pi * np.arange(self.n_angular // 2) / self.n_angular
@@ -273,7 +312,6 @@ class FLEBasis2D:
         x = x.reshape(1, -1)
         y = np.sin(phi)
         y = y.reshape(1, -1)
-        pts = pts.reshape(-1, 1)
         x = x * pts * h
         y = y * pts * h
         x = x.flatten()
@@ -325,6 +363,53 @@ class FLEBasis2D:
             a_conv = self.c2r @ (b * (self.r2c @ a).flatten())
 
         return a_conv.flatten()
+    
+    
+    def to_angular_order(self, a):
+        a_ordered = np.zeros_like(a)
+        blk_ind = self.blk_ind
+        idx_list = self.idx_list
+        for i in range(len(blk_ind) - 1):
+            idx_i = idx_list[i]
+            a_ordered[blk_ind[i]:blk_ind[i + 1]] = a[idx_i]
+
+        return a_ordered
+
+    def to_eigen_order(self, a_ordered):
+        a = np.zeros_like(a_ordered)
+        blk_ind = self.blk_ind
+        idx_list = self.idx_list
+        for i in range(len(blk_ind) - 1):
+            idx_i = idx_list[i]
+            a[idx_i] = a_ordered[blk_ind[i]:blk_ind[i + 1]]
+
+        return a
+
+
+    def expand_ctf(self, voltage_list, cs_list, alpha_list, defocus_list, pixel_size):
+        pts = self.pts
+        h = self.h
+        wavelength_list = 12.2643247 / np.sqrt(voltage_list * 1e3 + 0.978466 * voltage_list ** 2)
+        c2_vec = (-np.pi * wavelength_list * defocus_list).reshape(-1,1)
+        c4_vec = (0.5 * np.pi * (cs_list * 1e7) * wavelength_list ** 3).reshape(-1,1)
+
+        r2 = (pts * h / (pixel_size * 2 * np.pi)) ** 2
+        r4 = r2 ** 2
+        gamma = r2 @ c2_vec.T + r4 @ c4_vec.T
+        ctf_radial = np.sqrt(1 - alpha_list ** 2) * np.sin(gamma) - alpha_list * np.cos(gamma)
+
+        if self.n_interp > self.n_radial:
+            ctf_radial = dct(ctf_radial, axis=0, type=2) / (2 * self.n_radial)
+            ctf_radial_z = np.zeros(ctf_radial.shape)
+            ctf_radial = np.concatenate((ctf_radial, ctf_radial_z), axis=0)
+            ctf_radial = idct(ctf_radial, axis=0, type=2) * 2 * ctf_radial.shape[0]
+
+        rwts_mat = np.zeros((self.ne, len(voltage_list)), dtype=np.float64)
+
+        for i in range(self.ndmax + 1):
+            rwts_mat[self.idx_list[i], :] = self.A3[i] @ ctf_radial
+
+        return rwts_mat.T
 
     def rotate(self, a, theta):
 
